"""
grit.core
"""
import logging
import traceback
from typing import Any, Callable, Dict, Optional, Sequence, Type, Union

AnyException = Union[BaseException, Exception]

ExceptionType = Type[Exception]
# ExceptionType = Union[Type[Exception], Type[BaseException]]

GRIT_LOGGER = logging.getLogger("grit")


def format_traceback(ex: AnyException) -> str:
    """Generate a full exception traceback string from an exception."""
    tb_lines = traceback.format_exception(ex.__class__, ex, ex.__traceback__)
    return "".join(tb_lines)


def log_traceback(ex: AnyException, logger: logging.Logger, level: int):
    """Log the full exception traceback of an exception."""
    logger.log(level, format_traceback(ex))


# pylint: disable=fixme

# TODO: call async or standard sync handlers


class Grit:
    """Grit context manager for dealing with exceptions less badly."""

    def log_debug_traceback(self, ex: AnyException):
        """Log a full stacktrace at the DEBUG level."""
        self.logger.debug(format_traceback(ex))

    def __init__(
        self,
        dnr_list: Optional[Union[Sequence[AnyException], Sequence]] = None,
        handlers: Optional[Dict[ExceptionType, Callable[[AnyException], Any]]] = None,
        fallback_handler: Optional[Callable[[AnyException], Any]] = None,
        logger: logging.Logger = GRIT_LOGGER,
    ) -> None:
        self.dnr_list = tuple(dnr_list) if dnr_list else tuple()

        self.logger = logger
        self.handlers = handlers or {}
        self.exception: Optional[AnyException] = None
        self.fallback_handler = fallback_handler or self.log_debug_traceback
        self.result: Any = None

    def __enter__(self) -> "Grit":
        return self

    def __exit__(
        self,
        exc_type: Optional[ExceptionType],
        exc_value: Optional[AnyException],
        exc_tb,
    ):
        if exc_value:
            # TODO: make this work with exception child types
            # use the fallback_handler if no other handler is found
            exc_handler = self.handlers.get(exc_type, self.fallback_handler)  # type: ignore
            if exc_handler:
                # TODO: log the name of the handler and deal with unnamed functions (lambdas)
                self.logger.info(
                    f"Encountered {exc_type} handling with {exc_handler} ..."
                )
                # TODO: inspect to determine if handler takes arguments/no args and treat it
                # appropriately
                # https://docs.python.org/3/library/inspect.html
                self.result = exc_handler(exc_value)

        if isinstance(exc_value, self.dnr_list):
            # Returning False will cause the original exception to be propogated up
            # the stack and processed normally
            # https://docs.python.org/3/reference/datamodel.html#object.__exit__
            return False
        # Returning True swallows the original exception
        return True

    def __repr__(self) -> str:
        dnr_list = ", ".join((e.__name__ for e in self.dnr_list))
        handlers = ", ".join(
            (f"{e.__name__}: {fn.__name__}" for e, fn in self.handlers.items())
        )
        return (
            f"{self.__class__.__name__}(dnr_list=[{dnr_list}], handlers={{{handlers}}})"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['grit']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'py-grit',
    'version': '0.0.1a0',
    'description': 'Exception handling context manager.',
    'long_description': '# grit\n\n[![ci](https://github.com/Kilo59/grit/workflows/ci/badge.svg)](https://github.com/Kilo59/grit/actions)\n[![pypi version](https://img.shields.io/pypi/v/grit.svg)](https://pypi.org/project/grit/)\n![Python Versions](https://img.shields.io/pypi/pyversions/grit)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\nContext manager for shrugging off exceptions less badly.\n\nSubtitle: exception handling for lazy people.\n\n## Description and Rational\n\nDoes your code suffer from an overuse of bare exceptions?\n\nWhat if you could still shrug off exceptions when needed but not be so cringe about it?\n\n```python\ntry:\n    foobar()\nexcept Exception:\n    pass\n```\n\n```python\nfrom grit import Grit\n\n# Full exception stacktrace is automatically logged\nwith Grit():\n    foobar()\n```\n\n## Quick start\n\n```\npip install py-grit\n```\n\n```python\n>>> from grit import Grit\n>>> with Grit():\n...     raise RunTimeError("something bad")\n>>> print("Uh, everything is under control. Situation normal")\nUh, everything is under control. Situation normal\n\n```\n\nPropagate the errors you care about, while ignoring the ones you don\'t.\n\n```python\n>>> from grit import Grit\n>>> with Grit(dnr_list=[ZeroDivisionError]):\n...     42 / 0\nTraceback (most recent call last):\n    ...\nZeroDivisionError: division by zero\n\n```\n\nAnd handle those that require special attention\n\n```python\n>>> from grit import Grit\n>>> with Grit(handlers={ValueError: print}):\n...     raise ValueError("what have you done?!")\nwhat have you done?!\n\n```\n\n## Logging and results\n\n`Grit()` accepts a `fallback_handler` callable which will be called on the exception if no specific\n\'handler\' (`handlers`) is found.\n\nBy default the `fallback_handler` will log a full exception traceback at the debug level using `self.logger`.\n\nTo change this behavior, provide your own `fallback_handler` or explicitly set it a `None`.\n\n```python\n>>> from grit import Grit\n>>> with Grit(fallback_handler=print):\n...     raise TypeError("what have I done?!")\nwhat have I done?!\n\n```\n\n## Usage Examples\n\nTODO ...\n',
    'author': 'Gabriel Gore',
    'author_email': 'gabriel59kg@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kilo59/grit',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)
